 /**************************************************************************
  *                                                                        *
  *               Copyright (C) 1996, Silicon Graphics, Inc.               *
  *                          All Rights Reserved.                          *
  *                                                                        *
  *  The files in this subtree  contain  UNPUBLISHED  PROPRIETARY  SOURCE  *
  *  CODE of Silicon Graphics, Inc.;  the  contents  of  these  files may  *
  *  not be disclosed to third  parties,  copied  or  duplicated  in  any  *
  *  form, in whole or in part,  without  the  prior  written  permission  * 
  *  of  Silicon Graphics, Inc.                                            *
  *                                                                        *
  *  RESTRICTED RIGHTS LEGEND:                                             *
  *  Use,  duplication  or  disclosure  by  the  Government is subject to  *
  *  restrictions as set forth in subdivision (c)(1)(ii) of the Rights in  *
  *  Technical Data and Computer Software clause at  DFARS  252.227-7013,  *
  *  and/or in similar or successor  clauses in the FAR,  DOD or NASA FAR  *
  *  Supplement.  Unpublished - rights reserved  under the Copyright Laws  *
  *  of the United States.                                                 *
  *                                                                        *
  *  THIS SOFTWARE IS  PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,  *
  *  EXPRESS,  IMPLIED OR  OTHERWISE,  INCLUDING  WITHOUT LIMITATION, ANY  *
  *  WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.      *
  *                                                                        *
  *  IN  NO  EVENT  SHALL  SILICON  GRAPHICS  BE  LIABLE FOR ANY SPECIAL,  *
  *  INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF  ANY  KIND,  OR ANY  *
  *  DAMAGES WHATSOEVER RESULTING FROM LOSS  OF  USE,  DATA  OR  PROFITS,  *
  *  WHETHER OR NOT ADVISED OF  THE  POSSIBILITY  OF  DAMAGE,  AND ON ANY  *
  *  THEORY OF LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR  *
  *  PERFORMANCE OF THIS SOFTWARE.                                         *
  **************************************************************************/

/*
 *	geom -
 *		A software geometry package for computer graphics.
 *
 *				David J. Brown - 1983 
 */
#include <values.h>
#include <math.h>
#include <stdio.h>
#include <gl.h>

#define STACKSIZE	20

static float Vm_Stack[STACKSIZE][4][4];
static int   SP = -1;
static float Vm[4][4];	/* Viewing matrix */
static float Vsx, Vsy, Vcx, Vcy;	/* Viewport parameters */

/*
 * 	initmatstack - 
 *		Initialize the matrix stack and the viewport
 */
i_initmatstack()
{
    i_initmatrix();
    i_initviewport();
}

/*
 * 	initviewport - 
 *		Initialize the viewport
 */
i_initviewport()
{
    Vsx = 1.0;
    Vsy = 1.0;
    Vcx = 0.0;
    Vcy = 0.0;
}

/*
 * 	initmatrix - 
 *		Set the matrix to the identity	
 */
i_initmatrix()
{
    i_initmat(Vm);
}

i_initmat(matrix)
register float *matrix;
{
    *matrix++ = 1.0;	/* row 1	*/
    *matrix++ = 0.0;
    *matrix++ = 0.0;
    *matrix++ = 0.0;
    *matrix++ = 0.0;	/* row 2	*/
    *matrix++ = 1.0;
    *matrix++ = 0.0;
    *matrix++ = 0.0;
    *matrix++ = 0.0;	/* row 3	*/
    *matrix++ = 0.0;
    *matrix++ = 1.0;
    *matrix++ = 0.0;
    *matrix++ = 0.0;	/* row 4	*/
    *matrix++ = 0.0;
    *matrix++ = 0.0;
    *matrix++ = 1.0;
}

/*
 *  	viewport - 
 *		Set the current viewport
 */
i_viewport( l, r, b, t)
float l, r, b, t;
{
    Vsx = r-l;
    Vsy = t-b;
    Vcx = (r+l)/2.0;
    Vcy = (t+b)/2.0;
}

/*
 *  	pushmatrix - 
 *		Push the current matrix onto the stack
 */
i_pushmatrix()
{
    register int row, col;

    if (SP == STACKSIZE-1)
        puts("Stack is Full");
    else {
	SP++;
        for(row=0 ; row<4 ; row++)
            for(col=0 ; col<4 ; col++)
                Vm_Stack[SP][row][col] = Vm[row][col];
    }
}


/*
 *  	popmatrix - 
 *		Pop the current matrix off the stack
 */
i_popmatrix()
{
    register int row, col;

    if (SP == -1)
        puts("Stack is Empty");
    else {
        for(row=0 ; row<4 ; row++)
            for(col=0 ; col<4 ; col++)
                Vm[row][col] = Vm_Stack[SP][row][col];
	SP--;
    }
}

/*
 *	scale - 
 *		Scale the current matrix
 */
i_scale(Sx, Sy, Sz)
float Sx, Sy, Sz;
{
    register int i;

    for(i=0 ; i<4 ; i++)
	Vm[0][i] *= Sx;
    for(i=0 ; i<4 ; i++)
	Vm[1][i] *= Sy;
    for(i=0 ; i<4 ; i++)
	Vm[2][i] *= Sz;
}

/*
 *	loadmatrix - 
 *		Load the current matrix
 */
i_loadmatrix( m )
float m[4][4];
{
    register int i, j;

    for(i=0; i<4; i++) 
	for(j=0; j<4; j++) 
	    Vm[i][j] = m[i][j];
}

/*
 *	getmatrix - 
 *		Get the current matrix
 */
i_getmatrix( m )
float m[4][4];
{
    register int i, j;

    for(i=0; i<4; i++) 
	for(j=0; j<4; j++) 
	    m[i][j] = Vm[i][j];
}

/*
 *	translate - 
 *		Translate the current matrix
 */
i_translate(Tx, Ty, Tz)
float Tx, Ty, Tz;
{
    Vm[3][0] += (Tx*Vm[0][0] + Ty*Vm[1][0] + Tz*Vm[2][0]);
    Vm[3][1] += (Tx*Vm[0][1] + Ty*Vm[1][1] + Tz*Vm[2][1]);
    Vm[3][2] += (Tx*Vm[0][2] + Ty*Vm[1][2] + Tz*Vm[2][2]);
}

/*
 *	rotate - 
 *		Rotate about an axis; angle is in tenth of degrees
 */
i_rotate( angle, axis)
float angle;
char axis;
{
    register int row,col;
    float temp[4];
    float cosine, sine;

    for(col=0; col<4 ; col++)	/* init temp to zero matrix */
        temp[col] = 0;

    angle = angle*(3.1415926535/1800.0);
    cosine = cos(angle);
    sine = sin(angle);
    switch(axis){
    case 'x':    
    case 'X':    
        for(col=0 ; col<4 ; col++)
            temp[col] = cosine*Vm[1][col] + sine*Vm[2][col];
        for(col=0 ; col<4 ; col++) {
	    Vm[2][col] = - sine*Vm[1][col] + cosine*Vm[2][col];
            Vm[1][col] = temp[col];
	}
        break;

    case 'y':
    case 'Y':
        for(col=0 ; col<4 ; col++)
            temp[col] = cosine*Vm[0][col] - sine*Vm[2][col];
        for(col=0 ; col<4 ; col++) {
            Vm[2][col] = sine*Vm[0][col] + cosine*Vm[2][col];
            Vm[0][col] = temp[col];
        }
	break;

    case 'z':
    case 'Z':
        for(col=0 ; col<4 ; col++)
            temp[col] = cosine*Vm[0][col] + sine*Vm[1][col];
        for(col=0 ; col<4 ; col++) {
            Vm[1][col] = - sine*Vm[0][col] + cosine*Vm[1][col];
            Vm[0][col] = temp[col];
        }
	break;
    }
}

/* 
 *    	multmatrix - 
 *		Premultiply the current matrix 
 */
i_multmatrix( icand )
Matrix icand;
{
    register int row, col;
    float temp[4][4];

    for(row=0 ; row<4 ; row++) 
        for(col=0 ; col<4 ; col++)
            temp[row][col] = icand[row][0] * Vm[0][col]
                           + icand[row][1] * Vm[1][col]
                           + icand[row][2] * Vm[2][col]
                           + icand[row][3] * Vm[3][col];
    for(row=0 ; row<16 ; row++)
         *(Vm[0]+row) = *(temp[0]+row);
}

/*
 * 	transform -
 *		Transform the coordinates using the current matrix.
 */
i_transform( x, y, z, tx, ty, tz)	
float x, y, z, *tx, *ty, *tz;
{
    register int col;
    float Vc[4];

    for(col=0 ; col<3 ; col++)
        Vc[col] = x*Vm[0][col] + y*Vm[1][col] + z*Vm[2][col] + Vm[3][col];
    *tx = Vsx*Vc[0]+Vcx;
    *ty = Vsy*Vc[1]+Vcy;
    *tz = Vc[2];
}

/*
 * 	transform4 -
 *		Transform the coordinates using the current matrix.
 */
i_transform4(x,y,z,w,tx,ty,tz,tw)	
float x,y,z,w;
float *tx,*ty,*tz,*tw;
{
    register int col;
    float Vc[4];

    for(col=0 ; col<4 ; col++)
        Vc[col] = x*Vm[0][col] + y*Vm[1][col] + z*Vm[2][col] + w*Vm[3][col];
    *tx = Vsx*Vc[0]+Vcx;
    *ty = Vsy*Vc[1]+Vcy;
    *tz = Vc[2];
    *tw = Vc[3];
}


/*
 * 	polarview -
 *		Concatenate a polarview matrix 
 *
 */
i_polarview(dist, azimuth, incidence, twist)
float dist;
float azimuth, incidence, twist;
{
    i_translate(0.0, 0.0, -dist);
    i_rotate(-10.*twist,'z');	
    i_rotate(-10.*incidence,'x');
    i_rotate(-10.*azimuth,'z');
}

/*
 * 	lookat -
 *		Concatenate a lookat matrix 
 */
i_lookat( vx, vy, vz, px, py, pz, twist)
float vx, vy, vz, px, py, pz;
int twist;
{
    register float sine, cosine, hyp, hyp1, dx, dy, dz;
    register float mat[4][4];

    i_rotate(-10.*twist,'z');
    dx = px - vx;
    dy = py - vy;
    dz = pz - vz;
    hyp = dx * dx + dz * dz;	/* hyp squared	*/
    hyp1 = sqrt(dy*dy + hyp);
    hyp = sqrt(hyp);		/* the real hyp	*/
    i_initmat(mat);
    if (hyp1 != 0.0) {		/* rotate X	*/
	sine = -dy / hyp1;
	cosine = hyp /hyp1;
    } else {
	sine = 0;
	cosine = 1.0;
    }
    mat[1][1] = cosine;
    mat[1][2] = sine;
    mat[2][1] = -sine;
    mat[2][2] = cosine;
    i_multmatrix(mat);
    mat[1][1] = mat[2][2] = 1.0;	/* be careful here to reinit	*/
    mat[1][2] =	mat[2][1] = 0.0;	/* those modified by the last	*/
					/* paragraph	*/
    if (hyp != 0.0) {			/* rotate Y	*/
	sine = dx / hyp;
	cosine = -dz / hyp;
    } else {
	sine = 0;
	cosine = 1.0;
    }
    mat[0][0] = cosine;
    mat[0][2] = -sine;
    mat[2][0] = sine;
    mat[2][2] = cosine;
    i_multmatrix(mat);
    i_translate(-vx,-vy,-vz);	/* translate viewpoint to origin */
}

/*
 * 	window - 
 *		Set up a 3d PERSPECTIVE window. This loads the matrix onto
 * 	the stack destroying the current transformation.
 */
i_window(left, right, bottom, top, near, far)
float left, right, bottom, top, near, far;
{
    register float Xdelta, Ydelta, Zdelta;
    register float mat[4][4];

    Xdelta = right - left;
    Ydelta = top - bottom;
    Zdelta = far - near;
    if (Xdelta == 0.0 || Ydelta == 0.0 || Zdelta == 0.0) {
	fprintf(stderr,"window: window width, height, or depth is 0.0\n");
	return;
    }
    mat[0][0] = near * 2.0/Xdelta;
    mat[1][1] = near * 2.0/Ydelta;
    mat[2][0] = (right + left)/Xdelta;		/* note: negate Z	*/
    mat[2][1] = (top + bottom)/Ydelta;
    mat[2][2] = -(far + near)/Zdelta;
    mat[2][3] = -1.0;
    mat[3][2] = -2.0 * near * far / Zdelta;
    mat[0][1] = mat[0][2] = mat[0][3] =
    mat[1][0] = mat[1][2] = mat[1][3] =
    mat[3][0] = mat[3][1] = mat[3][3] = 0.0;
    i_loadmatrix(mat);
}

/*
 * 	perspective -
 *		Set up a perspective window. This loads the matrix onto 
 *	the stack destroying the current transformation.
 */
i_perspective(fovy,aspect,near,far)
float fovy, aspect, near, far;
{
    float Zdelta, cotangent;
    float matrix[4][4];

    if (fovy <= 0.1 || fovy > 180.0) {
 	fprintf(stderr,"perspective: fovy is out of range [0.1,180.0]\n");
	return;
    }
    Zdelta = far-near;
    if (aspect == 0.0 || Zdelta == 0.0) {
	fprintf(stderr,"perspective: window width or depth is 0.0\n");
	return;
    }
    fovy = ((fovy*M_PI)/180.0); 
    fovy /= 2;				/* take half the angle*/
    i_initmat(matrix);
    cotangent = cos(fovy) / sin(fovy);
    matrix[0][0] = cotangent / aspect;
    matrix[1][1] = cotangent;
    matrix[2][2] = -(far+near)/Zdelta;
    matrix[2][3] = -1.0;
    matrix[3][2] = -(2.0*far*near)/Zdelta;
    matrix[3][3] = 0.0;
    i_loadmatrix(matrix);
}

/*
 * 	ortho -
 *		Set up a 3d window. This loads the matrix onto the stack
 * 	destroying the current transformation.
 */
i_ortho(left, right, bottom, top, near, far)
float left, right, bottom, top, near, far;
{
    register float matrix[4][4];
    register float Xdelta, Ydelta, Zdelta;

    Xdelta = right - left;
    Ydelta = top - bottom;
    Zdelta = far - near;
    if (Xdelta == 0.0 || Ydelta == 0.0 || Zdelta == 0.0) {
	fprintf(stderr,"ortho: window width, height, or depth is 0.0\n");
	return;
    }
    i_initmat(matrix);
    matrix[0][0] = 2.0/Xdelta;
    matrix[3][0] = -(right + left)/Xdelta;
    matrix[1][1] = 2.0/Ydelta;
    matrix[3][1] = -(top + bottom)/Ydelta;
    matrix[2][2] = -2.0/Zdelta;		/* note: negate Z	*/
    matrix[3][2] = -(far + near)/Zdelta;
    i_loadmatrix(matrix);
}

/*
 * 	ortho2 -
 *		Set up a 2d window. This loads the matrix onto the stack
 * 	destroying the current transformation.  Note that the new 
 *	transformation leaves Z negated.
 */
i_ortho2(left, right, bottom, top)
float left, right, bottom, top;
{
    register float matrix[4][4];
    register float Xdelta, Ydelta;

    Xdelta = right - left;
    Ydelta = top - bottom;
    if (Xdelta == 0.0 || Ydelta == 0.0) {
	fprintf(stderr,"ortho2: window width or height is 0.0\n");
	return;
    }
    i_initmat(matrix);
    matrix[0][0] = 2.0/Xdelta;
    matrix[3][0] = -(right + left)/Xdelta;
    matrix[1][1] = 2.0/Ydelta;
    matrix[3][1] = -(top + bottom)/Ydelta;
    matrix[2][2] = -1.0;
    i_loadmatrix(matrix);
}


/*
 *	matrix -
 *		Some utilities for working with matricies.
 *
 *				Paul Haeberli - 1985
 */

invertmat(from,to)
float from[4][4],to[4][4];
{
    float wtemp[4][8];
    float m0,m1,m2,m3,s;
    float *r0,*r1,*r2,*r3, *rtemp;

    r0 = wtemp[0];
    r1 = wtemp[1];
    r2 = wtemp[2];
    r3 = wtemp[3];
    r0[0] = from[0][0];		/* build up [A][I]	*/
    r0[1] = from[0][1];
    r0[2] = from[0][2];
    r0[3] = from[0][3];
    r0[4] = 1.0;
    r0[5] = 0.0;
    r0[6] = 0.0;
    r0[7] = 0.0;
    r1[0] = from[1][0];
    r1[1] = from[1][1];
    r1[2] = from[1][2];
    r1[3] = from[1][3];
    r1[4] = 0.0;
    r1[5] = 1.0;
    r1[6] = 0.0;
    r1[7] = 0.0;
    r2[0] = from[2][0];
    r2[1] = from[2][1];
    r2[2] = from[2][2];
    r2[3] = from[2][3];
    r2[4] = 0.0;
    r2[5] = 0.0;
    r2[6] = 1.0;
    r2[7] = 0.0;
    r3[0] = from[3][0];
    r3[1] = from[3][1];
    r3[2] = from[3][2];
    r3[3] = from[3][3];
    r3[4] = 0.0;
    r3[5] = 0.0;
    r3[6] = 0.0;
    r3[7] = 1.0;

    if (r0[0] == 0.0) {		/* swap rows if needed		*/
	if (r1[0] == 0.0) {
	    if (r2[0] == 0.0) {
		if (r3[0] == 0.0) goto singular;
		rtemp = r0; r0 = r3; r3 = rtemp;
	    }
	    else {rtemp = r0; r0 = r2; r2 = rtemp;}
	}
	else {rtemp = r0; r0 = r1; r1 = rtemp;}
    }
    m1 = r1[0]/r0[0];		/* eliminate first variable	*/
    m2 = r2[0]/r0[0];
    m3 = r3[0]/r0[0];
    s = r0[1];
    r1[1] = r1[1] - m1 * s;
    r2[1] = r2[1] - m2 * s;
    r3[1] = r3[1] - m3 * s;
    s = r0[2];
    r1[2] = r1[2] - m1 * s;
    r2[2] = r2[2] - m2 * s;
    r3[2] = r3[2] - m3 * s;
    s = r0[3];
    r1[3] = r1[3] - m1 * s;
    r2[3] = r2[3] - m2 * s;
    r3[3] = r3[3] - m3 * s;
    s = r0[4];
    if (s != 0.0) {
	r1[4] = r1[4] - m1 * s;
	r2[4] = r2[4] - m2 * s;
	r3[4] = r3[4] - m3 * s;
    }
    s = r0[5];
    if (s != 0.0) {
	r1[5] = r1[5] - m1 * s;
	r2[5] = r2[5] - m2 * s;
	r3[5] = r3[5] - m3 * s;
    }
    s = r0[6];
    if (s != 0.0) {
	r1[6] = r1[6] - m1 * s;
	r2[6] = r2[6] - m2 * s;
	r3[6] = r3[6] - m3 * s;
    }
    s = r0[7];
    if (s != 0.0) {
	r1[7] = r1[7] - m1 * s;
	r2[7] = r2[7] - m2 * s;
	r3[7] = r3[7] - m3 * s;
    }

    if (r1[1] == 0.0) {		/* swap rows if needed		*/
	if (r2[1] == 0.0) {
	    if (r3[1] == 0.0) goto singular;
	    rtemp = r1; r1 = r3; r3 = rtemp;
	}
	else {rtemp = r1; r1 = r2; r2 = rtemp;}
    }
    m2 = r2[1]/r1[1];		/* eliminate second variable	*/
    m3 = r3[1]/r1[1];
    r2[2] = r2[2] - m2 * r1[2];
    r3[2] = r3[2] - m3 * r1[2];
    r3[3] = r3[3] - m3 * r1[3];
    r2[3] = r2[3] - m2 * r1[3];
    s = r1[4];
    if (s != 0.0) {
	r2[4] = r2[4] - m2 * s;
	r3[4] = r3[4] - m3 * s;
    }
    s = r1[5];
    if (s != 0.0) {
	r2[5] = r2[5] - m2 * s;
	r3[5] = r3[5] - m3 * s;
    }
    s = r1[6];
    if (s != 0.0) {
	r2[6] = r2[6] - m2 * s;
	r3[6] = r3[6] - m3 * s;
    }
    s = r1[7];
    if (s != 0.0) {
	r2[7] = r2[7] - m2 * s;
	r3[7] = r3[7] - m3 * s;
    }

    if (r2[2] == 0.0) {		/* swap last 2 rows if needed	*/
	if (r3[2] == 0.0) goto singular;
	rtemp = r2; r2 = r3; r3 = rtemp;
    }
    m3 = r3[2]/r2[2];		/* eliminate third variable	*/
    r3[3] = r3[3] - m3 * r2[3];
    r3[4] = r3[4] - m3 * r2[4];
    r3[5] = r3[5] - m3 * r2[5];
    r3[6] = r3[6] - m3 * r2[6];
    r3[7] = r3[7] - m3 * r2[7];

    if (r3[3] == 0.0) goto singular;
    s = 1.0/r3[3];		/* now back substitute row 3	*/
    r3[4] = r3[4] * s;
    r3[5] = r3[5] * s;
    r3[6] = r3[6] * s;
    r3[7] = r3[7] * s;

    m2 = r2[3];			/* now back substitute row 2	*/
    s = 1.0/r2[2];
    r2[4] = s * (r2[4] - r3[4] * m2);
    r2[5] = s * (r2[5] - r3[5] * m2);
    r2[6] = s * (r2[6] - r3[6] * m2);
    r2[7] = s * (r2[7] - r3[7] * m2);
    m1 = r1[3];
    r1[4] = (r1[4] - r3[4] * m1);
    r1[5] = (r1[5] - r3[5] * m1);
    r1[6] = (r1[6] - r3[6] * m1);
    r1[7] = (r1[7] - r3[7] * m1);
    m0 = r0[3];
    r0[4] = (r0[4] - r3[4] * m0);
    r0[5] = (r0[5] - r3[5] * m0);
    r0[6] = (r0[6] - r3[6] * m0);
    r0[7] = (r0[7] - r3[7] * m0);

    m1 = r1[2];			/* now back substitute row 1	*/
    s = 1.0/r1[1];
    r1[4] = s * (r1[4] - r2[4] * m1);
    r1[5] = s * (r1[5] - r2[5] * m1);
    r1[6] = s * (r1[6] - r2[6] * m1);
    r1[7] = s * (r1[7] - r2[7] * m1);
    m0 = r0[2];
    r0[4] = (r0[4] - r2[4] * m0);
    r0[5] = (r0[5] - r2[5] * m0);
    r0[6] = (r0[6] - r2[6] * m0);
    r0[7] = (r0[7] - r2[7] * m0);

    m0 = r0[1];			/* now back substitute row 0	*/
    s = 1.0/r0[0];
    r0[4] = s * (r0[4] - r1[4] * m0);
    r0[5] = s * (r0[5] - r1[5] * m0);
    r0[6] = s * (r0[6] - r1[6] * m0);
    r0[7] = s * (r0[7] - r1[7] * m0);

    to[0][0] = r0[4];		/* copy results back		*/
    to[0][1] = r0[5];
    to[0][2] = r0[6];
    to[0][3] = r0[7];
    to[1][0] = r1[4];
    to[1][1] = r1[5];
    to[1][2] = r1[6];
    to[1][3] = r1[7];
    to[2][0] = r2[4];
    to[2][1] = r2[5];
    to[2][2] = r2[6];
    to[2][3] = r2[7];
    to[3][0] = r3[4];
    to[3][1] = r3[5];
    to[3][2] = r3[6];
    to[3][3] = r3[7];
    return;

singular:
    fprintf(stderr,"invertmat: singular matrix\n");
    return;
}
