 /**************************************************************************
  *                                                                        *
  *               Copyright (C) 1996, Silicon Graphics, Inc.               *
  *                          All Rights Reserved.                          *
  *                                                                        *
  *  The files in this subtree  contain  UNPUBLISHED  PROPRIETARY  SOURCE  *
  *  CODE of Silicon Graphics, Inc.;  the  contents  of  these  files may  *
  *  not be disclosed to third  parties,  copied  or  duplicated  in  any  *
  *  form, in whole or in part,  without  the  prior  written  permission  * 
  *  of  Silicon Graphics, Inc.                                            *
  *                                                                        *
  *  RESTRICTED RIGHTS LEGEND:                                             *
  *  Use,  duplication  or  disclosure  by  the  Government is subject to  *
  *  restrictions as set forth in subdivision (c)(1)(ii) of the Rights in  *
  *  Technical Data and Computer Software clause at  DFARS  252.227-7013,  *
  *  and/or in similar or successor  clauses in the FAR,  DOD or NASA FAR  *
  *  Supplement.  Unpublished - rights reserved  under the Copyright Laws  *
  *  of the United States.                                                 *
  *                                                                        *
  *  THIS SOFTWARE IS  PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,  *
  *  EXPRESS,  IMPLIED OR  OTHERWISE,  INCLUDING  WITHOUT LIMITATION, ANY  *
  *  WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.      *
  *                                                                        *
  *  IN  NO  EVENT  SHALL  SILICON  GRAPHICS  BE  LIABLE FOR ANY SPECIAL,  *
  *  INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF  ANY  KIND,  OR ANY  *
  *  DAMAGES WHATSOEVER RESULTING FROM LOSS  OF  USE,  DATA  OR  PROFITS,  *
  *  WHETHER OR NOT ADVISED OF  THE  POSSIBILITY  OF  DAMAGE,  AND ON ANY  *
  *  THEORY OF LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR  *
  *  PERFORMANCE OF THIS SOFTWARE.                                         *
  **************************************************************************/

/*
    Author : Patrick Bouchaud
             galaad@neu.sgi.com
*/

#include <stdio.h>
#include <stdlib.h>
#include <X11/keysym.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <math.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/mman.h>

#include <GL/gl.h>
#include <GL/glu.h>
#include <GL/glx.h>
#include <GL/glut.h>

#include "mallocbis.h"
#include "manipulator.h"


static struct {
	int		lastx, lasty,
			middledown, rightdown, leftdown;
	float	dx, dy;
} mouse;

static float clipbox[] = {
	-1., 0., 0., .5,
	 1., 0., 0., .5,
	 0.,-1., 0., .5,
	 0., 1., 0., .5,
	 0., 0.,-1., .5,
	 0., 0., 1., .5,
};

static GLfloat
	PlaneOffset,
	MatRotPlane[16],
	MatRotCube[16],
	delta_roll = 0.,
	delta_pitch = 0.;

static int RotCube, RotPlane, newPlane=0, antialiased=1;

static GLdouble clipplane[] = { 0., 0., 1., 0. };

static void forceUpdate( void );
static void Redraw( void );
static void Mouse( int, int, int, int );
static void Motion( int, int );
static void Key( unsigned char, int, int );
static void Special( int, int, int );

/*---------------------------------------------------------------------------*/
/*	InitManipDisplay
/*---------------------------------------------------------------------------*/
int
InitManipDisplay( int x, int y, int w, int h )
{
	int i, win;
/*
	init to identity matrix
*/
	for( i=0; i<16; i++)
	{
		MatRotCube[i] =
		MatRotPlane[i] = (float) !(i%5);
	}

	glutInitDisplayMode( GLUT_RGB | GLUT_DOUBLE | GLUT_DEPTH );
	glutInitWindowPosition( x, y );
	glutInitWindowSize( w, h );
	win = glutCreateWindow( "manipulator" );
	glutDisplayFunc( Redraw );
	glutMouseFunc( Mouse );
	glutMotionFunc( Motion );
	glutKeyboardFunc( Key );
	glutSpecialFunc( Special );

	glMatrixMode( GL_PROJECTION );
	glLoadIdentity();
	gluPerspective( 45., 1., .1, 100. );
	glMatrixMode( GL_MODELVIEW );
	glLoadIdentity();

	glEnable( GL_DEPTH_TEST );
	glLineWidth( 3. );
	glEnable( GL_LINE_SMOOTH );
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ZERO);
	glClearColor( 0., 0., 0., 1. );
	glClearDepth( 1. );
	glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );
	glutSwapBuffers();

	Redraw();

	return win;
}

/*---------------------------------------------------------------------------*/
/*	Motion
/*---------------------------------------------------------------------------*/
static void
Motion( int x, int y )
{
	mouse.dx = x-mouse.lastx;
	mouse.dy = y-mouse.lasty;
	mouse.lastx = x;
	mouse.lasty = y;

	glutPostRedisplay();
}

/*---------------------------------------------------------------------------*/
/*	Special
/*---------------------------------------------------------------------------*/
static void
Special( int key, int x, int y )
{
	switch (key)
	{
		case GLUT_KEY_LEFT:
			SliceInterval *= .99;
			break;

		case GLUT_KEY_RIGHT:
			SliceInterval *= 1.01;
			break;

		case GLUT_KEY_UP:
			SliceNumber += 1.;
			break;

		case GLUT_KEY_DOWN:
			if (SliceNumber > 1.) SliceNumber -= 1.;
			break;
	}
	newPlane = GL_TRUE;
	updatePlaneStructure();
	glutPostRedisplay();
}

/*---------------------------------------------------------------------------*/
/*	Key
/*---------------------------------------------------------------------------*/
static void
Key( unsigned char key, int x, int y )
{
	switch (key)
	{
		case 27: /* Escape */
			exit(0);

		case 'a':
		case 'A':
			antialiased = !antialiased;
			if (antialiased)
			{
				glLineWidth( 3. );
				glEnable( GL_LINE_SMOOTH );
				glEnable(GL_BLEND);
				glBlendFunc(GL_SRC_ALPHA, GL_ZERO);
			}
			else
			{
				glDisable( GL_LINE_SMOOTH );
				glDisable(GL_BLEND);
				glLineWidth( 1. );
			}
			break;
	}
	glutPostRedisplay();
}

/*---------------------------------------------------------------------------*/
/*	Mouse
/*---------------------------------------------------------------------------*/
static void
Mouse( int button, int state, int x, int y )
{
	mouse.lastx = x;
	mouse.lasty = y;
	mouse.dx = 0.;
	mouse.dy = 0.;
	switch (button)
	{
		case GLUT_LEFT_BUTTON:
			mouse.leftdown = (state==GLUT_DOWN);
			break;

		case GLUT_RIGHT_BUTTON:
			mouse.rightdown = (state==GLUT_DOWN);
			break;

		case GLUT_MIDDLE_BUTTON:
			mouse.middledown = (state==GLUT_DOWN);
			break;
	}
	glutPostRedisplay();
}

/*---------------------------------------------------------------------------*/
/*			updateMatrices
/*---------------------------------------------------------------------------*/
static void
updateMatrices()
{
	float invmat[4][4], equation[4];

	RotCube = GL_FALSE;
	RotPlane = GL_FALSE;
	delta_roll  = (float) mouse.dx;
	delta_pitch = (float) mouse.dy;
	if (mouse.leftdown)
	{
		if (mouse.middledown)
		{
			RotCube = GL_TRUE;
			RotPlane = GL_TRUE;
		}
		else
		{
			PlaneOffset += ((float) mouse.dy)/100.;
			newPlane = GL_TRUE;
		}
	}
	else if (mouse.middledown)
	{
		RotPlane = GL_TRUE;
		newPlane = GL_TRUE;
	}

	if ( RotCube )
	{
		glPushMatrix();
		glLoadIdentity();
		glRotatef( delta_roll,  0., 1., 0. );
		glRotatef( delta_pitch, 1., 0., 0. );
		glMultMatrixf( MatRotCube );
		glGetFloatv( GL_MODELVIEW_MATRIX, MatRotCube );
		glPopMatrix();
	}

	if ( RotPlane )
	{
		glPushMatrix();
		glLoadIdentity();
		glRotatef( delta_roll,  0., 1., 0. );
		glRotatef( delta_pitch, 1., 0., 0. );
		glMultMatrixf( MatRotPlane );
		glGetFloatv( GL_MODELVIEW_MATRIX, MatRotPlane );
		glPopMatrix();
	}

/*
	Compute plane equation used by the voxel view
*/
	invertmat( MatRotCube, invmat );
	i_initviewport();
	i_loadmatrix(invmat);
	i_multmatrix(MatRotPlane);
	i_transform( 0., 0., 1., equation, equation+1, equation+2 );
	equation[3] = PlaneOffset;
	Plane.Equation[0] = (GLdouble) equation[0];
	Plane.Equation[1] = (GLdouble) equation[1];
	Plane.Equation[2] = (GLdouble) equation[2];
	Plane.Equation[3] = (GLdouble) equation[3];

	updatePlaneStructure();

	mouse.dx = mouse.dy = 0.;
}

/*---------------------------------------------------------------------------*/
/*	Redraw
/*---------------------------------------------------------------------------*/
static void
Redraw( void )
{
	updateMatrices();

	glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );
	glPushMatrix();
	glTranslatef( 0., 0., -2.5 );

/*
	Draw over clipplane
	-------------------
*/
	glPushMatrix();
	glMultMatrixf( MatRotCube );
	glClipPlane( GL_CLIP_PLANE0, Plane.Equation);
	glEnable( GL_CLIP_PLANE0 );
	glPopMatrix();

	glPushMatrix();
	glMultMatrixf( MatRotCube );
	glColor3f( 1., 1., 1. );
	drawaxis();
	glColor3f( 0., 1., 0. );
	drawcube();
	glPopMatrix();

/*
	Draw under clipplane
	--------------------
*/
	Plane.Equation[0] *= -1.;
	Plane.Equation[1] *= -1.;
	Plane.Equation[2] *= -1.;
	Plane.Equation[3] *= -1.;
	glPushMatrix();
	glMultMatrixf( MatRotCube );
	glClipPlane( GL_CLIP_PLANE0, Plane.Equation);
	glPopMatrix();
	Plane.Equation[0] *= -1.;
	Plane.Equation[1] *= -1.;
	Plane.Equation[2] *= -1.;
	Plane.Equation[3] *= -1.;

	glPushMatrix();
	glMultMatrixf( MatRotCube );
	glColor3f( 0., 0., 1. );
	drawaxis();
	drawcube();
	glPopMatrix();

/*
	Draw clipping plane
	-------------------
*/
	glDisable( GL_CLIP_PLANE0 );

	glPushMatrix();
	glMultMatrixf( MatRotCube );
	glColor3f( 0., 0., 1. );
	drawplanes();
	glPopMatrix();

	glPopMatrix();
	glutSwapBuffers();

	delta_roll = 0.;
	delta_pitch = 0.;

	if (newPlane)
	{
		newPlane = GL_FALSE;
		glutIdleFunc( forceUpdate );
	}
}

/*---------------------------------------------------------------------------*/
/*			forceUpdate
/*---------------------------------------------------------------------------*/
static void
forceUpdate( void )
{
	glutSetWindow( voxelWin );
	glutPostRedisplay();
	glutIdleFunc( NULL );
}

/*---------------------------------------------------------------------------*/
/*			drawaxis
/*---------------------------------------------------------------------------*/
drawaxis()
{
	glBegin( GL_LINES );
		glVertex3f(-.5, 0., 0.);
		glVertex3f( .5, 0., 0.);
		glVertex3f( 0.,-.5, 0.);
		glVertex3f( 0., .5, 0.);
		glVertex3f( 0., 0.,-.5);
		glVertex3f( 0., 0., .5);
	glEnd();
	glBegin( GL_TRIANGLE_FAN );
		glVertex3f(.5, 0., 0.);
		glVertex3f(.45, .02, 0.);
		glVertex3f(.45, 0., .02);
		glVertex3f(.45,-.02, 0.);
		glVertex3f(.45, 0.,-.02);
		glVertex3f(.45, .02, 0.);
	glEnd();
	glBegin( GL_TRIANGLE_FAN );
		glVertex3f(0., .5, 0.);
		glVertex3f(.02, .45, 0.);
		glVertex3f(0., .45, .02);
		glVertex3f(-.02, .45, 0.);
		glVertex3f(0., .45, -.02);
		glVertex3f(.02, .45, 0.);
	glEnd();
	glBegin( GL_TRIANGLE_FAN );
		glVertex3f(0., 0., .5);
		glVertex3f(0., .02, .45);
		glVertex3f(.02, 0., .45);
		glVertex3f(0., -.02, .45);
		glVertex3f(-.02, 0., .45);
		glVertex3f(0., .02, .45);
	glEnd();
}

/*---------------------------------------------------------------------------*/
/*			drawcube
/*---------------------------------------------------------------------------*/
drawcube()
{
 glBegin( GL_LINE_LOOP );
	glVertex3f(-.5,-.5,-.5);
	glVertex3f(-.5, .5,-.5);
	glVertex3f( .5, .5,-.5);
	glVertex3f( .5,-.5,-.5);
 glEnd();

 glBegin( GL_LINE_LOOP );
	glVertex3f(-.5,-.5,-.5);
	glVertex3f(-.5,-.5, .5);
	glVertex3f( .5,-.5, .5);
	glVertex3f( .5,-.5,-.5);
 glEnd();

 glBegin( GL_LINE_LOOP );
	glVertex3f(-.5,-.5,-.5);
	glVertex3f(-.5, .5,-.5);
	glVertex3f(-.5, .5, .5);
	glVertex3f(-.5,-.5, .5);
 glEnd();

 glBegin( GL_LINE_LOOP );
	glVertex3f(-.5, .5,-.5);
	glVertex3f(-.5, .5, .5);
	glVertex3f( .5, .5, .5);
	glVertex3f( .5, .5,-.5);
 glEnd();

 glBegin( GL_LINE_LOOP );
	glVertex3f( .5,-.5,-.5);
	glVertex3f( .5,-.5, .5);
	glVertex3f( .5, .5, .5);
	glVertex3f( .5, .5,-.5);
 glEnd();

 glBegin( GL_LINE_LOOP );
	glVertex3f(-.5,-.5, .5);
	glVertex3f( .5,-.5, .5);
	glVertex3f( .5, .5, .5);
	glVertex3f(-.5, .5, .5);
 glEnd();
}

/*---------------------------------------------------------------------------*/
/*			drawplanes
/*---------------------------------------------------------------------------*/
/*
	Drawing - from farest to nearest slice
*/
int
drawplanes()
{
	int i, j, numVertex;
	GLfloat centre[3];
	Vertex *vertex[4], *v;

	centre[0] = Plane.farCentre[0];
	centre[1] = Plane.farCentre[1];
	centre[2] = Plane.farCentre[2];
	for (i=0; i<SliceNumber; i++)
	{
		centre[0] -= Plane.sliceInterval[0];
		centre[1] -= Plane.sliceInterval[1];
		centre[2] -= Plane.sliceInterval[2];

		for (j=0; j<4; j++)
		{
			newVertex( vertex[j] );
		}
		for (j=0; j<4; j++)
		{
			vertex[j]->last = (j==0) ? vertex[3] : vertex[j-1];
			vertex[j]->next = (j==3) ? vertex[0] : vertex[j+1];
		}

#define setVertex( v, corner )\
		v->x = corner(0); v->y = corner(1); v->z = corner(2);

		setVertex( vertex[0], LB );
		setVertex( vertex[1], LT );
		setVertex( vertex[2], RT );
		setVertex( vertex[3], RB );

		numVertex = 4;
		for (j=0; j<6; j++) cutplane( &vertex[0], &numVertex, &clipbox[4*j] );

		if (numVertex)
		{
			glBegin( GL_LINE_STRIP );
			for (j=0, v=vertex[0]; j<=numVertex; j++, v=v->next)
				glVertex3f( v->x, v->y, v->z );
			glEnd();
		}

		for (j=0, v=vertex[0]; j<numVertex; j++)
		{
			Vertex *nextv = v->next;
			freeVertex( v );
			v = nextv;
		}
	}
}

/*---------------------------------------------------------------------------*/
/*			cutplane
/*---------------------------------------------------------------------------*/
#define	equate(v)\
	( equation[0]*v->x + equation[1]*v->y + equation[2]*v->z + equation[3])

/*
    compute the 3D intersection of the
        line: (v1,v2)
        line: eq[0]*v->x + eq[1]*v->y + eq[2]*v->z + eq[3] = 0
    (purely mathematical)
*/
static Vertex *
ClipLine3D( Vertex *v1, Vertex *v2, float *equation )
{
    float t;
    Vertex *v;

	newVertex(v);
    t = equate(v1);
    if ( t == 0. )
    {
        v->x = v1->x;
        v->y = v1->y;
        v->z = v1->z;
    }
    else
    {
        t = t/(t-equate(v2));
        v->x = v1->x + t*( v2->x - v1->x );
        v->y = v1->y + t*( v2->y - v1->y );
        v->z = v1->z + t*( v2->z - v1->z );
    }
    return v;
}


int
cutplane( Vertex **polygon, int *numVertex, float *equation )
{
	Vertex *v = *polygon, *v0, *v1, *v2, *v3;
	int num = *numVertex;

	if (v == NULL) return;

	if (equate(v) >= 0)	/* first vertex in */
	{
		do v = v->next;
		while ( (v==*polygon) ? 0 : (equate(v)>=0) );

		if (v==*polygon)	/* all vertice in */
			return;

		v0 = v->last;
		v1 = ClipLine3D( v->last, v, equation );

		do v = v->next;
		while ( equate(v) < 0 );

		v2 = ClipLine3D( v->last, v, equation );
		v3 = v;

		for (v=v->last; v!=v0; )
		{
			v=v->last;
			freeVertex( v->next );
			num--;
		}
		v0->next = v1;
		v1->last = v0;
		v1->next = v2;
		v2->last = v1;
		v2->next = v3;
		v3->last = v2;

		*numVertex = num+2;
		*polygon = v0;
		return;
	}
	else /* first vertex out */
	{
		do v = v->next;
		while ( (v==*polygon) ? 0 : (equate(v)<0) );

		if (v==*polygon)	/* all vertice out */
		{
			do
			{
				Vertex *nextv = v->next;
				freeVertex( v );
				v = nextv;
			}
			while (v!=*polygon);
			*polygon = NULL;
			*numVertex = 0;
			return;
		}

		v0 = ClipLine3D( v->last, v, equation );
		v1 = v;

		do v = v->next;
		while ( equate(v) >= 0 );

		v2 = v->last;
		v3 = ClipLine3D( v->last, v, equation );

		while ( v!=v1 )
		{
			v=v->next;
			freeVertex( v->last );
			num--;
		}
		v0->last = v3;
		v0->next = v1;
		v1->last = v0;
		v2->next = v3;
		v3->last = v2;
		v3->next = v0;

		*numVertex = num+2;
		*polygon = v0;
		return;
	}
}

/*---------------------------------------------------------------------------*/
/*			updatePlaneStructure
/*---------------------------------------------------------------------------*/
/*
	Knowing clipping plane equation, we compute :
	---------------------------------------------

	the lookup vector Y onto that plane :
	Y = projection of x onto that plane, i.e.
		x = a N + b Y, thus Y = [x - (x.N) N]/[normalized]

	the complementary vector X onto that plane :
	X = normal ^ up
	
	the centre C of coordinates onto that plane :
	C = projection of O onto the plane, along its normal
*/
int
updatePlaneStructure()
{
	int	i;
	GLfloat	up[3], cp[3];

/*
	lookup vector (plane normal is supposed to be normalized)
*/
	if (Plane.Equation[0] == 1.)
	{
		up[0] = 0.;
		up[1] = 1.;
		up[2] = 0.;
	}
	else if (Plane.Equation[2] > 0.)
	{
		up[0] = 1. - Plane.Equation[0]*Plane.Equation[0];
		up[1] =    - Plane.Equation[0]*Plane.Equation[1];
		up[2] =    - Plane.Equation[0]*Plane.Equation[2];
		NRMVEC3( up );
	}
	else
	{
		up[0] = -1.+ Plane.Equation[0]*Plane.Equation[0];
		up[1] =    + Plane.Equation[0]*Plane.Equation[1];
		up[2] =    + Plane.Equation[0]*Plane.Equation[2];
		NRMVEC3( up );
	}

/*
	complementary vector
*/
	cp[0] = Plane.Equation[1]*up[2] - Plane.Equation[2]*up[1];
	cp[1] = Plane.Equation[2]*up[0] - Plane.Equation[0]*up[2];
	cp[2] = Plane.Equation[0]*up[1] - Plane.Equation[1]*up[0];

/*
	Centre of coordinates (plane normal is supposed to be normalized)
		- compute centre of farest slice -
*/
	SETVEC3( Plane.farCentre, Plane.Equation );
	SCLVEC3( Plane.farCentre, -Plane.Equation[3]-SliceNumber*SliceInterval );

	SETVEC3( Plane.sliceInterval, Plane.Equation );
	SCLVEC3( Plane.sliceInterval, -SliceInterval );

	ADDVEC3( Plane.up, cp, up );
	SUBVEC3( Plane.dn, cp, up );

	SETVEC3( Plane.S, cp );
	SETVEC3( Plane.T, up );

#if 0
	SCLVEC3( Plane.up, 1.7321 );
	SCLVEC3( Plane.dn, 1.7321 );
#endif
}
