 /**************************************************************************
  *                                                                        *
  *               Copyright (C) 1996, Silicon Graphics, Inc.               *
  *                          All Rights Reserved.                          *
  *                                                                        *
  *  The files in this subtree  contain  UNPUBLISHED  PROPRIETARY  SOURCE  *
  *  CODE of Silicon Graphics, Inc.;  the  contents  of  these  files may  *
  *  not be disclosed to third  parties,  copied  or  duplicated  in  any  *
  *  form, in whole or in part,  without  the  prior  written  permission  * 
  *  of  Silicon Graphics, Inc.                                            *
  *                                                                        *
  *  RESTRICTED RIGHTS LEGEND:                                             *
  *  Use,  duplication  or  disclosure  by  the  Government is subject to  *
  *  restrictions as set forth in subdivision (c)(1)(ii) of the Rights in  *
  *  Technical Data and Computer Software clause at  DFARS  252.227-7013,  *
  *  and/or in similar or successor  clauses in the FAR,  DOD or NASA FAR  *
  *  Supplement.  Unpublished - rights reserved  under the Copyright Laws  *
  *  of the United States.                                                 *
  *                                                                        *
  *  THIS SOFTWARE IS  PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,  *
  *  EXPRESS,  IMPLIED OR  OTHERWISE,  INCLUDING  WITHOUT LIMITATION, ANY  *
  *  WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.      *
  *                                                                        *
  *  IN  NO  EVENT  SHALL  SILICON  GRAPHICS  BE  LIABLE FOR ANY SPECIAL,  *
  *  INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF  ANY  KIND,  OR ANY  *
  *  DAMAGES WHATSOEVER RESULTING FROM LOSS  OF  USE,  DATA  OR  PROFITS,  *
  *  WHETHER OR NOT ADVISED OF  THE  POSSIBILITY  OF  DAMAGE,  AND ON ANY  *
  *  THEORY OF LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR  *
  *  PERFORMANCE OF THIS SOFTWARE.                                         *
  **************************************************************************/

/*
    Author : Patrick Bouchaud
             galaad@neu.sgi.com
*/

#include <stdio.h>
#include <stdlib.h>
#include <X11/keysym.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <math.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/mman.h>

#include <GL/gl.h>
#include <GL/glu.h>
#include <GL/glx.h>
#include <GL/glut.h>

#include "mallocbis.h"
#include "manipulator.h"

#define	TILE_XSIZE	128
#define	TILE_YSIZE	128
#define	TILE_ZSIZE	64

#define	DIMX	256
#define	DIMY	256
#define	DIMZ	128

static Tile *firstTile = NULL;
static GLubyte voxel[DIMX*DIMY*DIMZ];

static GLfloat zoomFactor = .9;
static int	interpolated = 1,
			textured = 1,
			blendMode = 1,
			newImage = 1;

static struct {
    int     lastx, lasty,
            middledown, rightdown, leftdown;
    float   dx, dy;
} mouse;

static int W, H, maxcoord[3], dimx, dimy, dimz;

static void Reshape( int w, int h );
static void Redraw( void );
static void Mouse( int, int, int, int );
static void Motion( int, int );
static void Key( unsigned char, int, int );
static void Special( int, int, int );

static void tileDefine( Tile ** );


/*---------------------------------------------------------------------------*/
/*	InitVoxelDisplay
/*---------------------------------------------------------------------------*/
int
InitVoxelDisplay( int x, int y, int w, int h )
{
	int win, i, fd;
	GLubyte *data;
	GLubyte *voxelptr;
	struct stat status;

#ifdef DB
	glutInitDisplayMode( GLUT_RGB | GLUT_DOUBLE );
#else
	glutInitDisplayMode( GLUT_RGB | GLUT_SINGLE );
#endif
	glutInitWindowPosition( x, y );
	glutInitWindowSize( w, h );
	win = glutCreateWindow( "voxel" );
	glutDisplayFunc( Redraw );
	glutReshapeFunc( Reshape );
	glutMouseFunc( Mouse );
	glutMotionFunc( Motion );
	glutKeyboardFunc( Key );

	fd = open( "../data/256_256_124.bin", O_RDONLY );
	fstat( fd, &status );
	data =
	(GLubyte *) mmap(NULL,(size_t)status.st_size,PROT_READ,MAP_SHARED,fd,0);

	voxelptr = voxel;
	for (i=0; i<status.st_size; i++)
	{
		*voxelptr++ = *data++;
	}
	close( fd );

/*
	create the linked list of all tiles parameters
	----------------------------------------------
*/
	dimx = DIMX;
	dimy = DIMY;
	dimz = DIMZ;
	tileDefine( &firstTile );

	return win;
}

/*---------------------------------------------------------------------------*/
/*	InitGraphique
/*---------------------------------------------------------------------------*/
static void
InitGraphique( void )
{
	glBlendFunc( GL_ONE, GL_ZERO );
	if (SliceNumber > 1.)
	{
		glBlendEquationEXT( GL_MAX_EXT );
		glEnable( GL_BLEND );
	}
	else
	{
		glDisable( GL_BLEND );
	}

	glMatrixMode( GL_PROJECTION );
	glLoadIdentity();
	glOrtho( -.5, .5, -.5, .5, -1., 1. );
	glMatrixMode( GL_MODELVIEW );
	glLoadIdentity();

	glClearColor( 0., 0., 0., 1. );
	glClear( GL_COLOR_BUFFER_BIT );
#ifdef DB
	glutSwapBuffers();
#else
	glFlush();
#endif
}

/*---------------------------------------------------------------------------*/
/*	Reshape
/*---------------------------------------------------------------------------*/
static void
Reshape( int w, int h )
{
	static int first = 1;
	glViewport( 0, 0, w, h );
	W = w;
	H = h;
	if (first)
	{
		first = 0;	
		InitGraphique();
	}
}

/*---------------------------------------------------------------------------*/
/*	Mouse
/*---------------------------------------------------------------------------*/
static void Mouse( int button, int state, int x, int y )
{
	mouse.lastx = x;
	mouse.lasty = y;
	mouse.dx = 0.;
	mouse.dy = 0.;
	switch (button)
	{
		case GLUT_LEFT_BUTTON:
			mouse.leftdown = (state==GLUT_DOWN);
			break;

		case GLUT_RIGHT_BUTTON:
			mouse.rightdown = (state==GLUT_DOWN);
			break;

		case GLUT_MIDDLE_BUTTON:
			mouse.middledown = (state==GLUT_DOWN);
			break;
	}
}

/*---------------------------------------------------------------------------*/
/*	Motion
/*---------------------------------------------------------------------------*/
static void
Motion( int x, int y )
{
	mouse.dx = x-mouse.lastx;
	mouse.dy = y-mouse.lasty;
	mouse.lastx = x;
	mouse.lasty = y;

	glutPostRedisplay();
}

/*---------------------------------------------------------------------------*/
/*	Key
/*---------------------------------------------------------------------------*/
static void
Key( unsigned char key, int x, int y )
{
	switch (key)
	{
		case 27:	/* Escape */
			exit(0);

		case 'T':
		case 't': textured = !textured; break;

		case 'M':
		case 'm': blendMode++; break;

		case 'i':
		case 'I': interpolated = !interpolated; break;
	}
	glutPostRedisplay();
}

/*---------------------------------------------------------------------------*/
/*	tileDefine
/*---------------------------------------------------------------------------*/
static void
tileDefine( Tile **first )
{
	Tile	*tile;
	int		dim, xndx, yndx, zndx, first_one;

	maxcoord[0] = dimx-1;
	maxcoord[1] = dimy-1;
	maxcoord[2] = dimz-1;

	first_one = 1;
	newTile( *first );
	tile = *first;
	tile->id = 0;
	tile->last = NULL;
	for (xndx=0; xndx+TILE_XSIZE<=dimx; xndx+=(TILE_XSIZE-1))
	{
		for (yndx=0; yndx+TILE_YSIZE<=dimy; yndx+=(TILE_YSIZE-1))
		{
			for (zndx=0; zndx+TILE_ZSIZE<=dimz; zndx+=(TILE_ZSIZE-1))
			{
				if (!first_one)
				{
					newTile( tile->next );
					tile->next->last = tile;
					tile = tile->next;
					tile->next = NULL;
				}
				else first_one = 0;
				
/*
				define the 6 inequations of the faces of the tile
				we set the equations, so that for each dimension,
				in pixel coordinates :
					orig+.5 <= pixelcoord <= (orig+size-.5)
				and
					we map it to [-.5,.5]
*/
				tile->size[0] = TILE_XSIZE;
				tile->size[1] = TILE_YSIZE;
				tile->size[2] = TILE_ZSIZE;

				tile->orig[0] = xndx;
				tile->orig[1] = yndx;
				tile->orig[2] = zndx;

				for (dim=0; dim<3; dim++)
				{
					float *eq = &tile->equation[8*dim];
					eq[0] = eq[1] = eq[2] = eq[3] = 0.;
					eq[4] = eq[5] = eq[6] = eq[7] = 0.;

					eq[dim] = 1.;
					eq[3] = -(float)tile->orig[dim]-.5;

					eq[dim+4] = -1.;
					eq[7] = (float)(tile->orig[dim]+tile->size[dim])-.5;

					eq[3] = (eq[3]/(float)maxcoord[dim])+.5;
					eq[7] = (eq[7]/(float)maxcoord[dim])-.5;
				}
			}
		}
	}
}

/*---------------------------------------------------------------------------*/
/*	windowLevelIt
/*---------------------------------------------------------------------------*/
#ifdef WL_16_BITS
#define	MAXWIDTH	65536
#define	WL_TYPE		GL_UNSIGNED_SHORT
#define	WL_DW		mouse.dx*100
#define	WL_DL		mouse.dy*100
typedef u_short		WLtype;
#else
#define	MAXWIDTH	256
#define	WL_TYPE		GL_UNSIGNED_BYTE
#define	WL_DW		mouse.dx
#define	WL_DL		mouse.dy
typedef u_char		WLtype;
#endif

static void
windowLevelIt( void )
{
	float scale, bias;
	static int width=MAXWIDTH, level=MAXWIDTH/2;
	static WLtype WLbuffer[1280*1024];

	glDisable( GL_TEXTURE_3D_EXT );

	if (newImage)
	{
		newImage = GL_FALSE;
		glReadBuffer( GL_BACK );
/*
		glReadPixels( ..., GL_LUMINANCE, ... ) would
		read back the SUM of r,g,b component (cf man page)
*/
		glReadPixels( 0, 0, W, H, GL_RED, WL_TYPE, WLbuffer);
	}

	if (mouse.middledown)
	{
		width += WL_DW;
		level -= WL_DL;
	}
#if 0
	fprintf( stderr, "width=%d, level=%d\n", width, level );
#endif

/*
	compute the scale & bias coefficients from :
		result[ pixel ] = pixel*scale + bias;
		result[ (level-width/2)/MAXWIDTH ] = 0.
		result[ (level+width/2)/MAXWIDTH ] = 1.
*/
	scale = (float)MAXWIDTH/(float)width;
	bias = .5 - (float)level/(float)width;

	glPixelTransferf(GL_RED_SCALE,   scale);
	glPixelTransferf(GL_GREEN_SCALE, scale);
	glPixelTransferf(GL_BLUE_SCALE,  scale);
	glPixelTransferf(GL_RED_BIAS,    bias);
	glPixelTransferf(GL_GREEN_BIAS,  bias);
	glPixelTransferf(GL_BLUE_BIAS,   bias);

	glBlendFunc( GL_ONE, GL_ZERO );
	glBlendEquationEXT( GL_FUNC_ADD_EXT );
	glDisable( GL_BLEND );
	glDrawPixels( W, H, GL_LUMINANCE, WL_TYPE, WLbuffer );

#ifdef DB
	glutSwapBuffers();
#else
	glFlush();
#endif

	glPixelTransferf(GL_RED_SCALE,   1.);
	glPixelTransferf(GL_GREEN_SCALE, 1.);
	glPixelTransferf(GL_BLUE_SCALE,  1.);
	glPixelTransferf(GL_RED_BIAS,    0.);
	glPixelTransferf(GL_GREEN_BIAS,  0.);
	glPixelTransferf(GL_BLUE_BIAS,   0.);
}

/*---------------------------------------------------------------------------*/
/*	setTexEnvironment
/*---------------------------------------------------------------------------*/
static void
setTexEnvironment()
{
	if (!textured)
	{
		glDisable( GL_TEXTURE_3D_EXT );
		glDisable( GL_BLEND );
		glPolygonMode( GL_FRONT_AND_BACK, GL_LINE );
		glClear( GL_COLOR_BUFFER_BIT );
		return;
	}

	glPolygonMode( GL_FRONT_AND_BACK, GL_FILL );
/*
	Blending
*/
	if (SliceNumber == 1.)
		glDisable( GL_BLEND );
	else switch (blendMode%2)
	{
		case 0:
			glBlendEquationEXT( GL_MIN_EXT );
			glEnable( GL_BLEND );
			glClearColor( 1., 1., 1., 1. );
			break;

		case 1:
			glBlendEquationEXT( GL_MAX_EXT );
			glEnable( GL_BLEND );
			glClearColor( 0., 0., 0., 1. );
			break;

		case 2:
			glBlendEquationEXT( GL_FUNC_ADD_EXT );
#if 0
			glBlendColorEXT( 1., 1., 1., 1./(float)SliceNumber );
			glBlendFunc( GL_CONSTANT_ALPHA_EXT, GL_ONE );
#else
			glBlendFunc( GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA );
#endif
			glEnable( GL_BLEND );
			glClearColor( 0., 0., 0., 0. );
			break;
	}

	glClear( GL_COLOR_BUFFER_BIT );
	glEnable( GL_TEXTURE_3D_EXT );
}

/*---------------------------------------------------------------------------*/
/*	loadTile
/*---------------------------------------------------------------------------*/
static void
loadTile( Tile *tile )
{
	if (tile->id)
	{
		glBindTextureEXT( GL_TEXTURE_3D_EXT, tile->id );
	}
	else
	{
		glGenTexturesEXT( 1, &tile->id );
		glBindTextureEXT( GL_TEXTURE_3D_EXT, tile->id );
/*
	indicate unpacking parameters
*/
		glPixelStorei(GL_UNPACK_ROW_LENGTH, dimx);
		glPixelStorei(GL_UNPACK_IMAGE_HEIGHT_EXT, dimy);
		glPixelStorei(GL_UNPACK_SKIP_PIXELS, tile->orig[0] );
		glPixelStorei(GL_UNPACK_SKIP_ROWS, tile->orig[1] );
		glPixelStorei(GL_UNPACK_SKIP_IMAGES_EXT, tile->orig[2] );
/*
		load texture tile
*/
		glTexImage3DEXT(
			GL_TEXTURE_3D_EXT,
			0,
			GL_LUMINANCE8_EXT,
			tile->size[0],
			tile->size[1],
			tile->size[2],
			0,
			GL_LUMINANCE,
			GL_UNSIGNED_BYTE,
			voxel
		);

		glPixelStorei(GL_UNPACK_ROW_LENGTH, 0);
		glPixelStorei(GL_UNPACK_IMAGE_HEIGHT_EXT, 0);
		glPixelStorei(GL_UNPACK_SKIP_PIXELS, 0);
		glPixelStorei(GL_UNPACK_SKIP_ROWS, 0);
		glPixelStorei(GL_UNPACK_SKIP_IMAGES_EXT, 0);
	}
/*
	Texturing parameters
*/
	glTexParameterf(GL_TEXTURE_3D_EXT, GL_TEXTURE_WRAP_S, GL_CLAMP);
	glTexParameterf(GL_TEXTURE_3D_EXT, GL_TEXTURE_WRAP_T, GL_CLAMP);
	glTexParameterf(GL_TEXTURE_3D_EXT, GL_TEXTURE_WRAP_R_EXT, GL_CLAMP);
	if (interpolated)
	{
		glTexParameterf(GL_TEXTURE_3D_EXT,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
		glTexParameterf(GL_TEXTURE_3D_EXT,GL_TEXTURE_MIN_FILTER,GL_LINEAR);
	}
	else
	{
		glTexParameterf(GL_TEXTURE_3D_EXT,GL_TEXTURE_MAG_FILTER,GL_NEAREST);
		glTexParameterf(GL_TEXTURE_3D_EXT,GL_TEXTURE_MIN_FILTER,GL_NEAREST);
	}

	glTexEnvf( GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE );

/*
	transform coordinates so that it belongs to [0,1]
*/
	glMatrixMode( GL_TEXTURE );
	glLoadIdentity();
	glScalef(
		(float)maxcoord[0]/(float)tile->size[0],
		(float)maxcoord[1]/(float)tile->size[1],
		(float)maxcoord[2]/(float)tile->size[2]
	);
	glTranslatef(
		.5-(float)tile->orig[0]/(float)maxcoord[0],
		.5-(float)tile->orig[1]/(float)maxcoord[1],
		.5-(float)tile->orig[2]/(float)maxcoord[2]
	);
	glMatrixMode( GL_MODELVIEW );
	glLoadIdentity();
	glRotatef( 270., 0., 0., 1. );
	glScalef(zoomFactor, zoomFactor, zoomFactor );

}

/*---------------------------------------------------------------------------*/
/*	Redraw
/*---------------------------------------------------------------------------*/
/*
	Drawing - from furthest to nearest slice
*/
static void
Redraw( void )
{
	int	i, j, numVertex;
	GLfloat centre[3];
	Tile *tile;
	Vertex *vertex[4], *v;

	if (mouse.middledown)
	{
		windowLevelIt();
		return;
	}

	newImage = GL_TRUE;
	if (mouse.leftdown)
	{
		zoomFactor += ((float) mouse.dy)/100.;
	}

	setTexEnvironment();

	for (tile=firstTile; tile!=NULL; tile=tile->next)
	{
		int tile_loaded = 0;
		glColor4f( 1., 1., 1., 1./(float)SliceNumber );
		centre[0] = Plane.farCentre[0];
		centre[1] = Plane.farCentre[1];
		centre[2] = Plane.farCentre[2];
		for (i=0; i<SliceNumber; i++)
		{
			for (j=0; j<4; j++)
			{
				newVertex( vertex[j] );
			}
			for (j=0; j<4; j++)
			{
				vertex[j]->last = (j==0) ? vertex[3] : vertex[j-1];
				vertex[j]->next = (j==3) ? vertex[0] : vertex[j+1];
			}
/*
			compute vertex coordinates of the slice
			---------------------------------------
*/
#define setVertex(v,corner)\
			v->x = corner(0); v->y = corner(1); v->z = corner(2);

			centre[0] -= Plane.sliceInterval[0];
			centre[1] -= Plane.sliceInterval[1];
			centre[2] -= Plane.sliceInterval[2];

			setVertex( vertex[0], LB );
			setVertex( vertex[1], LT );
			setVertex( vertex[2], RT );
			setVertex( vertex[3], RB );
			numVertex = 4;
/*
			3D-texture coordinates are computed as
			--------------------------------------
			intersection of the slice with the tile
*/
	        for (j=0; j<6; j++)
				cutplane( &vertex[0], &numVertex, &tile->equation[4*j] );

			if (numVertex)
			{
				if ( textured && !tile_loaded)
				{
					loadTile( tile );
					tile_loaded = 1;
				}
				glBegin( GL_POLYGON );
				for (j=0, v=vertex[0]; j<=numVertex; j++, v=v->next)
				{
					float vc[3];
/*
					2D vertex coordinates are computed as
					-------------------------------------
					projection of the vertex onto the "clipping" plane

					x = DOTPROD3( (v-centre), Plane.S );
					y = DOTPROD3( (v-centre), Plane.T );
*/
					vc[0] = v->x-centre[0];
					vc[1] = v->y-centre[1];
					vc[2] = v->z-centre[2];
					glTexCoord3f( v->x, v->y, v->z );
					glVertex2f( DOTPROD3(vc,Plane.S), DOTPROD3(vc,Plane.T));
				}
				glEnd();
			}

			for (j=0, v=vertex[0]; j<numVertex; j++)
			{
				Vertex *nextv = v->next;
				freeVertex( v );
				v = nextv;
			}
		}
	}

	windowLevelIt();
}
